/* This file is part of Legendes II */
/* Copyright (c) 1998 by
 *   Philippe Clavel (philippe.clavel@supelec.fr)
 *   Benjamin Lerman (benjamin.lerman@ens.fr)
 *   David Madore (david.madore@ens.fr) */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* File listener.c */

/* Thus speaketh ye mighty CVS:
 * $Id: listener.c,v 1.2 1999/05/02 23:21:36 david Exp $ */

/* This file contains the function that opens the initial socket and
 * the main function for the listener thread. */

/* CVS identity of this file. */
static char id[] = "$Id: listener.c,v 1.2 1999/05/02 23:21:36 david Exp $";

/* Include files */

#ifndef _THREAD_SAFE
#define _THREAD_SAFE		/* Make sure libc is reentrant. */
#endif

#include <stdio.h>		/* Standard I/O */
#include <stdlib.h>		/* Standard library functions */

#include <errno.h>		/* Error numbers */

#include <sys/types.h>		/* System types */
#include <sys/socket.h>		/* Socket types */
#include <sys/time.h>		/* For struct timeval */
#include <unistd.h>		/* UNIX standard functions */
#include <fcntl.h>		/* For fcntl() and its constants */

#include <netinet/in.h>		/* INET domain (TCP/IP) sockets */

#include <netdb.h>		/* gethostbyaddr() and friends */

#include <pthread.h>		/* Posix threads */


#include "server.h"

/* Main functions */

void
init_gsock(void)
     /* Create the initial TCP/IP socket, bind() it to port PORTNUM,
      * and start listen()ing to it.  This function is exectuted in
      * single thread mode. */
{
  struct sockaddr_in name; /* To store socket ``name'' in */

  /* Create socket. */
  if ((gsock=socket(AF_INET,SOCK_STREAM,0))<0)
    {
      perror("Can't create socket()");
      exit(1);
    }
  /* Make it non-blocking.  See NOTE0006 for why we need this. */
  if ( fcntl (gsock, F_SETFL, O_NONBLOCK) < 0 )
    {
      perror("Can't make socket non-blocking");
      exit(1);
    }
  /* Bind it to all interfaces, TCP port PORTNUM. */
  memset(&name,0,sizeof(name));
  name.sin_family = AF_INET;
  name.sin_port = htons(PORTNUM); /* TCP port number for our server */
  name.sin_addr.s_addr = htonl (INADDR_ANY); /* Accept all interfaces */
  if (bind(gsock,(struct sockaddr *)&name,sizeof(name))<0)
    {
      perror("Can't bind() socket");
      exit(1);
    }
  /* Start listen()ing. */
  if (listen(gsock,10)<0)
    {
      perror("Can't listen() to socket");
      exit(1);
    }
}

void
listener_loop(void)
     /* This function performs a loop that accepts incoming
      * connections and receives data from already established
      * connections. */
{
  fd_set fs;
  int s;
  struct sockaddr_in other;
  int other_namelen;
#if (DEBUG_LEVEL>=3)
  struct hostent *other_host;
#endif /* DEBUG_LEVEL>=3 */
  int max;
  int len;
  int i;
  struct timeval tv;
  struct lsock_s *sck;
  long prior_conn_count;
  char suresel;

  while (1)
    {
      /* First we select() - that is, we wait for something to
       * happen. */
#if (DEBUG_LEVEL>=5)
      fprintf(stderr,"[%0#10lx]Waiting for something to happen...\n",
	      pthread_self());
#endif /* DEBUG_LEVEL>=5 */
      /* Describe what we want to wait for.  We put every local
       * socket's desciptor in the fd_set for select() (as for max, it
       * contains the highest value of a file descriptor in the set).
       * Of course, we must hold the lsock_mutex to do all this. */
      max = gsock;
      FD_ZERO(&fs);
      FD_SET(gsock,&fs);
      safe_pthread_mutex_lock(&lsock_mutex);
      prior_conn_count = conn_count;
      for (i=0;i<lsock_nb;i++)
	{
	  if ( lsock_tab[i]->sopen )
	    {
	      if ( lsock_tab[i]->s > max )
		max = lsock_tab[i]->s;
	      FD_SET (lsock_tab[i]->s, &fs);
	    }
	}
      safe_pthread_mutex_unlock(&lsock_mutex);
      /* [NOTE0002] We have now released the lsock_mutex (of course we
       * can't afford to block on select() while holding such a
       * precious resource!).  This means that the lsock_tab may have
       * changed a lot when we emerge from our sleep.  It may even
       * change between the time when we release the lsock_mutex and
       * the time we start the select() (i.e. while you are reading
       * this :-).  It is even possible for some file desciptors from
       * the table to be closed, or even reused in that time.  This is
       * not a serious problem (we just use select() as an indication
       * of what we should do, not as an authoritative information),
       * but we should be aware of this fact. */
      /* Now do the waiting. */
      tv.tv_sec = 2; /* Timeout */
      tv.tv_usec = 0;
      if ( ( i = select (max+1, &fs, NULL, NULL, &tv) ) < 0 )
	{
	  if ( errno == EBADF )
	    /* This means that some file descriptor was no longer
	     * valid by the time we reached the select() (see
	     * NOTE0002).  This is not serious.  We just ignore and
	     * redo the select(). */
	    {
#if (DEBUG_LEVEL>=4)
	      fprintf(stderr,"[%0#10lx]Somebody stole my file descriptor :-(\n",
		      pthread_self());
#endif /* DEBUG_LEVEL>=4 */
	      sched_yield();
	      continue;
	    }
	  perror("select()ing");
	  exit(1);
	}
      if (i==0)
	/* Nothing happened (i.e. select() timed out.  We start
         * again. */
	{
	  sched_yield();
	  continue;
	}
      /* Examine what we have found. */
      /* [NOTE0004] While it is true that select() will never wake up
       * more than one sleeping thread for the same file descriptor,
       * yet we can't use that fact, because our threads are not
       * necessarily all sleeping at the same time.  So thread A may
       * be awoken for file descriptor f by select(), and before it
       * has time to take care of f, thread B may start select()ing
       * and be also awoken for the same descriptor.  So one of the
       * two is going to find out, by the time it gets to the file
       * descriptor, that the latter has nothing to say after all,
       * despite select() having indicated it as ready.  So even if
       * the file descriptor is set in fs, we must check that there
       * really is something to be found there.  An even more devilish
       * problem, see NOTE0002, is that a file descriptor may have the
       * same number as the one we selected on but actually refer to
       * something completely different (the former one has been
       * closed and the OS has reused the descriptor number).  So,
       * caveat auditor. */
      if (FD_ISSET(gsock,&fs))
	/* Global socket has something to say => new connection
         * possibly received */
	{
#if (DEBUG_LEVEL>=3)
	  fprintf(stderr,"[%0#10lx]Accepting new connection.\n",
		  pthread_self());
#endif /* DEBUG_LEVEL>=3 */
	  other_namelen = sizeof(other);
	  if ( ( s = accept (gsock, &other, &other_namelen) ) < 0 )
	    {
	      if ( errno == EAGAIN )
		/* [NOTE0006] As indicated in NOTE0004, we don't
		 * necessarily have a connection.  Here, someone stole
		 * it from us.  This is why we had to make gsock
		 * non-blocking (otherwise, the thread would have
		 * started sleeping and it would have essentially
		 * died). */
		{
#if (DEBUG_LEVEL>=4)
		  fprintf(stderr,"[%0#10lx]Somebody stole my connection :-(\n",
			  pthread_self());
#endif /* DEBUG_LEVEL>=4 */
		  sched_yield();
		  continue;
		}
	      perror("Can't accept()");
	      exit(1);
	    }
#if (DEBUG_LEVEL>=3)
	  fprintf(stderr,"[%0#10lx]Received connection:\n",
		  pthread_self());
	  fprintf(stderr,"[%0#10lx]Port number %d\n",
		  pthread_self(),other.sin_port);
	  /* XXX - gethostbyaddr() is not reentrant.  We should use
	   * gethostbyaddr_r() instead, shouldn't we?  Ugh... */
	  if ((other_host=gethostbyaddr((const char *)&other.sin_addr,
					sizeof(other.sin_addr),
					AF_INET))==NULL)
	    {
	      herror("Can't resolve address");
	    }
	  else
	    {
	      fprintf(stderr,"[%0#10lx]Host name: %s\n",
		      pthread_self(),other_host->h_name);
	    }
#endif /* DEBUG_LEVEL>=3 */
	  /* Add this new connection to the list of connections. */
	  safe_pthread_mutex_lock(&lsock_mutex);
	  if (((lsock_tab = realloc(lsock_tab,
				    (++lsock_nb)*sizeof(struct lsock_s *)))
	       /* realloc() the lsock_tab to make space for a new
                * pointer */
	       ==NULL)||
	      ((lsock_tab[lsock_nb-1] = malloc(sizeof(struct lsock_s)))
	       /* And allocate something at the end of this new
                * pointer */
	       ==NULL))
	    {
	      perror("Out of memory");
	      exit(1);
	    }
	  /* Advertise */
	  conn_count++; /* lsock_mutex is held */
	  /* Now initialize the data entries */
	  safe_pthread_mutex_init(&lsock_tab[lsock_nb-1]->d_mutex,NULL);
	  lsock_tab[lsock_nb-1]->sopen = 1;
	  lsock_tab[lsock_nb-1]->s = s;
	  lsock_tab[lsock_nb-1]->remote = other;
	  lsock_tab[lsock_nb-1]->handwork = 1;
	  lsock_tab[lsock_nb-1]->state = CONN_STATE_NEW;
	  lsock_tab[lsock_nb-1]->buflen = 0;
#if (DEBUG_LEVEL>=3)
	  fprintf(stderr,"[%0#10lx]We now have %d open connections.\n",
		  pthread_self(),lsock_nb);
#endif /* DEBUG_LEVEL>=3 */
	  /* Now release the lock.  !!! Do not do this earlier, for we
	   * use lsock_nb! */
	  safe_pthread_mutex_unlock(&lsock_mutex);
	  /* Wake up the handler threads.  Do this *after* the
           * lsock_mutex has been unlocked to avoid sleeping with the
           * lsock_mutex locked (see NOTE0008). */
	  safe_pthread_mutex_lock (&handler_wake_mutex);
	  handler_wake_count++;
	  pthread_cond_signal (&handler_wake_cond);
	  safe_pthread_mutex_unlock (&handler_wake_mutex);
	}
      else
	/* Data received on some socket */
	{
#if (DEBUG_LEVEL>=4)
	  fprintf(stderr,"[%0#10lx]Received data on some socket.\n",
		  pthread_self());
#endif /* DEBUG_LEVEL>=4 */
	  /* Find out which socket spoke */
	  sck=NULL;
	  safe_pthread_mutex_lock(&lsock_mutex);
	  if ( conn_count != prior_conn_count )
	    /* Some connection has been accepted or closed while we
	     * weren't paying attention.  I.e. we're out of date.
	     * It's better to give up and redo the select().
	     * Actually, it wouldn't really hurt to remove this whole
	     * test, but I think it's safer like that. */
	    {
	      safe_pthread_mutex_unlock(&lsock_mutex);
#if ( DEBUG_LEVEL >= 3 )
	      fprintf(stderr,"[%0#10lx]Warning: overwhelmed by novelty.\n",
		      pthread_self());
#endif /* DEBUG_LEVEL >= 3 */
	      sched_yield();
	      continue;
	    }
	  for ( i=0 ; i<lsock_nb ; i++)
	    {
	      if ( ( lsock_tab[i]->sopen )
		   && ( FD_ISSET (lsock_tab[i]->s, &fs) ) )
		/* Found the socket which spoke. */
		{
		  sck = lsock_tab[i];
		  break;
		}
	    }
	  if (!sck)
	    /* None did.  This is strange.  I think this shouldn't
	     * happen (this should be caught by the conn_count test
	     * above).  If it does, it is probably a misunderstanding
	     * on my part of what select() does. */
	    {
	      safe_pthread_mutex_unlock(&lsock_mutex);
#if (DEBUG_LEVEL>=1)
	      fprintf(stderr,"[%0#10lx]Warning: select() selected nothing.\n",
		      pthread_self());
#endif /* DEBUG_LEVEL>=1 */
	      sched_yield();
	      continue;
	    }
	  if (safe_pthread_mutex_trylock(&sck->d_mutex))
	    /* ...trylock() and not ...lock(), because we're not
             * allowed to block while owning the lsock_mutex (this
             * thing is _precious_, see NOTE0001). */
	    {
	      safe_pthread_mutex_unlock(&lsock_mutex);
#if (DEBUG_LEVEL>=3)
	      fprintf(stderr,"[%0#10lx]Warning: can't acquire d_mutex, giving up.\n",
		      pthread_self());
#endif /* DEBUG_LEVEL>=3 */
	      sched_yield();
	      continue;
	    }
	  safe_pthread_mutex_unlock(&lsock_mutex);
	  /* As pointed out (see NOTE0001), we don't need to hold
	   * lsock_mutex any longer to access the data in sck, now
	   * that we have the d_mutex.  And on the other hand, we
	   * could not release this sooner because we didn't have the
	   * d_mutex yet. */
	  if ( ! sck->sopen )
	    /* [NOTE0005] Someone has already closed the connection.
	     * So we have nothing to do, and we do it. */
	    {
	      safe_pthread_mutex_unlock(&sck->d_mutex);
#if (DEBUG_LEVEL>=3)
	      fprintf(stderr,"[%0#10lx]Warning: connection already obsoleted.\n",
		      pthread_self());
#endif /* DEBUG_LEVEL>=3 */
	      sched_yield();
	      continue;
	    }
	  s = sck->s; /* The socket descriptor */
	  len = BUFSIZE - sck->buflen; /* How much we try to read */
	  if ( len == 0 )
	    /* Buffer full.  We should give the handler threads some
	     * time to digest. */
	    {
	      safe_pthread_mutex_unlock(&sck->d_mutex);
#if (DEBUG_LEVEL>=3)
	      fprintf(stderr,"[%0#10lx]Warning: write buffer full.\n",
		      pthread_self());
#endif /* DEBUG_LEVEL>=3 */
	      sched_yield();
	      continue;
	    }
	  suresel = 0;
	suresel_label:
	  if ( ( i = read (s, &(sck->buffer[sck->buflen]), len) ) < 0 )
	    {
	      perror("read()ing from socket");
	      exit(1);
	    }
	  if ( i==0 )
	    /* [NOTE0003] Memento: when a socket is shut down by the
	     * other party, select() always finds something
	     * interesting on that socket, but a read() returns 0.
	     * This is the only way to know a socket has shut down,
	     * apparently (.vomit.).  So getting 0 as a result of
	     * read() is suspicious.  But we can only be sure that the
	     * socket was actually closed if the select() has been
	     * performed with the d_mutex held, because otherwise some
	     * other thread may have simply snatched the data from us.
	     * The suresel variable indicates whether we have
	     * performed a select() with the d_mutex held.  So when we
	     * read 0 (i.e. now), either suresel is false in which
	     * case we make it true (by doing the select() in
	     * question) and retry, or it is already true and then we
	     * can safely close the socket.  Disgusting?  You bet.
	     * Blame it on the Unix / pthreads interface (``seamless''
	     * is not precisely the word which comes to my mind). */
	    {
	      if ( suresel )
		/* No doubt about it, the connection was closed. */
		{
#if (DEBUG_LEVEL>=3)
		  fprintf (stderr, "[%0#10lx]Closing a connection.\n",
			   pthread_self ());
#endif /* DEBUG_LEVEL>=3 */
		  sck->sopen = 0;
		  shutdown (s, 2);
		  close (s);
		  safe_pthread_mutex_unlock (&sck->d_mutex);
		  /* Of course, release the d_mutex before locking the
		   * lsock_mutex! */
		  safe_pthread_mutex_lock (&lsock_mutex);
		  conn_count++; /* While holding the lsock_mutex */
		  safe_pthread_mutex_unlock (&lsock_mutex);
		  /* Wake up the handler threads, see NOTE0008. */
		  safe_pthread_mutex_lock (&handler_wake_mutex);
		  handler_wake_count++;
		  pthread_cond_signal (&handler_wake_cond);
		  safe_pthread_mutex_unlock (&handler_wake_mutex);
		  sched_yield ();
		  continue;
		}
	      else
		/* We aren't *certain* that the connection has closed.
		 * We must perform another select() operation, with
		 * zero timeout, and *while* holding the d_mutex. */
		{
		  FD_ZERO (&fs);
		  FD_SET (s,&fs);
		  tv.tv_sec = 0; /* Zero timeout */
		  tv.tv_usec = 0;
		  if ( ( i = select (max+1, &fs, NULL, NULL, &tv) ) < 0 )
		    {
		      perror ("select()ing");
		      exit(1);
		    }
		  if ( i == 1 )
		    /* Now we are sure that select() selects the socket.
		     * So we can retry the read(). */
		    {
		      suresel = 1;
		      goto suresel_label;
		      /* Yucky?  Yup.  Totally wirth-less.  ;-) */
		    }
		  else
		    /* False alarm. */
		    {
		      safe_pthread_mutex_unlock(&sck->d_mutex);
#if (DEBUG_LEVEL>=4)
		      fprintf (stderr, "[%0#10lx]Data snatched away from me.\n",
			       pthread_self ());
#endif /* DEBUG_LEVEL>=4 */
		    }
		  sched_yield();
		  continue;
		}
	    }
	  /* At this point, we have really read data from the
           * socket. */
#if (DEBUG_LEVEL>=4)
	  fprintf(stderr,"[%0#10lx]Read %d bytes from socket.\n",
		  pthread_self(),i);
#endif /* DEBUG_LEVEL>=4 */
	  sck->buflen += i;
	  sck->handwork = 1;
	  safe_pthread_mutex_unlock(&sck->d_mutex);
	  /* Wake up the handler threads.  Do this *after* the d_mutex
           * has been unlocked to avoid sleeping with the d_mutex
           * locked, see NOTE0008. */
	  safe_pthread_mutex_lock (&handler_wake_mutex);
	  handler_wake_count++;
	  pthread_cond_signal (&handler_wake_cond);
	  safe_pthread_mutex_unlock (&handler_wake_mutex);
	}
    }
}
